from .params import ReserveParams
from pydantic.main import BaseModel
from ....messages.types import  RESERVE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional



class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    with_progress: Optional[bool] = False

class MetaModel(MessageMetaModel):
    type: str = RESERVE
    extensions: Optional[MetaExtensionsModel]

class DataModel(MessageDataModel):
    node: Optional[str] #TODO: Maybe not optional
    template: Optional[str]
    provision: Optional[str]
    params: Optional[ReserveParams]


class ReserveMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from typing import Union
from bergen.debugging import DebugLevel
from bergen.handlers.base import ContractHandler
from bergen.messages import BouncedUnassignMessage, UnassignDoneMessage, UnassignCriticalMessage
from bergen.console import console


class UnassignHandler(ContractHandler[BouncedUnassignMessage]):

    async def log(self, message, level=DebugLevel.INFO):
        await self.connection.forward(message)

    @property
    def meta(self):
        return {"extensions": self.message.meta.extensions, "reference": self.message.meta.reference}

    async def pass_progress(self, text: str = None, percentage: int = None):
        console.log(text + percentage)

    async def pass_done(self):
        error_message = UnassignDoneMessage(data={"assignation": self.message.data.assignation}, meta=self.meta)
        await self.forward(error_message)

    async def pass_exception(self, message: BouncedUnassignMessage, exception: Union[Exception,str]):
        error_message = UnassignCriticalMessage(data={"message": str(exception), "type": exception.__class__.__name__}, meta=self.meta)
        await self.forward(error_message)



#Base Types
EXCEPTION = "exception"
ALLOWANCE = "allowance"

ASSIGNATION = "assignation"
PROVISION = "provision"
ASSIGNATION_REQUEST = "assignation_request"
PROVISION_REQUEST = "provision_request"

ACTIVATE_POD = "activate_pod"
DEACTIVATE_POD ="deactivate_pod"

# ASIGN PATH
ASSIGN = "assign"
BOUNCED_ASSIGN = "bounced_assign"
BOUNCED_FORWARDED_ASSIGN = "bounced_forwarded_assign"

ASSIGN_DONE = "assign_done"
ASSIGN_CANCELLED = "assign_cancelled"
ASSIGN_LOG ="assign_log"
ASSIGN_CRITICAL = "assign_fail"
ASSIGN_YIELD =  "assign_yield"
ASSIGN_RETURN = "assign_return"
ASSIGN_RETRY = "assign_retry"

# UNASSIGN PATH
UNASSIGN = "unassign"
BOUNCED_UNASSIGN = "bounced_unassign"

UNASSIGN_DONE = "unassign_done"
UNASSIGN_LOG ="unassign_log"
UNASSIGN_CRITICAL = "unassign_fail"

# Reserve
RESERVE = "reserve"
BOUNCED_RESERVE = "bounced_reserve"
BOUNCED_FORWARDED_RESERVE = "bounced_forwarded_reserve"

RESERVE_DONE = "reserve_done"
RESERVE_LOG = "reserve_log"
RESERVE_FAIL = "reserve_fail"
RESERVE_TRANSITION = "reserve_transition"
RESERVE_CRITICAL = "reserve_critical"
RESERVE_ACTIVE = "reserve_active"
RESERVE_WAITING = "reserve_waiting"

#Unreserve
UNRESERVE = "unreserve"
BOUNCED_UNRESERVE = "bounced_unreserve"

UNRESERVE_DONE = "unreserve_done"
UNRESERVE_LOG = "unreserve_log"
UNRESERVE_FAIL = "unreserve_fail"
UNRESERVE_CRITICAL = "unreserve_critical"


# Provide
PROVIDE = "provide"
BOUNCED_PROVIDE = "bounced_provide"

PROVIDE_DONE = "provide_done"
PROVIDE_LOG = "provide_log"
PROVIDE_TRANSITION = "provide_transition"
PROVIDE_FAIL = "provide_fail"
PROVIDE_CRITICAL = "provide_critical"
# Unprovide
UNPROVIDE = "unprovide"
BOUNCED_UNPROVIDE = "bounced_unprovide"

UNPROVIDE_DONE = "unprovide_done"
UNPROVIDE_LOG = "unprovide_log"
UNPROVIDE_FAIL = "unprovide_fail"
UNPROVIDE_CRITICAL = "unprovide_critical"
UNPROVIDE_ERROR = "unprovide_error"





from bergen.types.node.widgets.intwidget import IntWidget
from bergen.types.node.ports.returns.base import BaseReturnPort


class IntReturnPort(BaseReturnPort):

  def __init__(self, **kwargs) -> None:
      super().__init__(**kwargs)



from bergen.queries.delayed.node import DETAIL_NODE_FR
from bergen.schema import Assignation, Node, Peasent, Template, Pod, Provision, Transcript, VartPod, Volunteer
from bergen.query import TypedGQL



 
# Peasent Constants

SERVE_GQL = TypedGQL("""
    mutation Serve($name: String!){
        serve(name: $name){
            id
            name
            
        }
    }
""", Peasent)



OFFER_GQL = TypedGQL("""
    mutation Offer($node: ID!, $params: GenericScalar!, $policy: GenericScalar!){
        offer(node: $node, params: $params, policy: $policy){
            id
            name
            policy 
        }
    }
""", Template)


ACCEPT_GQL = TypedGQL("""
    mutation Accept($template: ID!, $provision: String!){
        accept(template: $template, provision: $provision){
            id
            template {
                node {"""
                + DETAIL_NODE_FR +
                """

                }
            }
        }
    }
""", Pod)





from bergen.actors.base import Actor
from bergen.handlers import *
from bergen.utils import *
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from bergen.actors.utils import *

import asyncio

class FunctionalActor(Actor):
    pass


class FunctionalFuncActor(FunctionalActor):
    
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def assign(self, *args, **kwargs):
        raise NotImplementedError("Please provide a func or overwrite the assign method!")

    async def _assign(self, assign_handler: AssignHandler, args, kwargs):
    
        assign_handler_context.set(assign_handler)
        provide_handler_context.set(self.provide_handler)
        #
        result = await self.assign(*args, **kwargs)

        try:
            shrinked_returns = await shrinkOutputs(self.template.node, result) if self.shrinkOutputs else result
            await assign_handler.pass_result(shrinked_returns)
        except Exception as e:
            await assign_handler.pass_exception(e)


class FunctionalGenActor(FunctionalActor):

    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def assign(self,*args, **kwargs):
        raise NotImplementedError("This needs to be overwritten in order to work")

    async def _assign(self, assign_handler: AssignHandler, args, kwargs):

        assign_handler_context.set(assign_handler)
        provide_handler_context.set(self.provide_handler)

        try:
            async for result in self.assign(*args, **kwargs):
                lastresult = await shrinkOutputs(self.template.node, result) if self.shrinkOutputs else result
                await assign_handler.pass_yield(lastresult)

            await assign_handler.pass_done()
        except Exception as e:
            await assign_handler.pass_exception(e)


class FunctionalThreadedFuncActor(FunctionalActor):
    nworkers = 5

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPoolExecutor(self.nworkers)

    def assign(self, *args, **kwargs):
        raise NotImplementedError("")   


    def _assign_threaded(self, args, kwargs, loop, assign_handler, provide_handler):
        assign_handler_context.set(assign_handler)
        provide_handler_context.set(provide_handler)
        loop_context.set(loop)
        res = self.assign(*args, **kwargs)
        loop_context.set(None)
        return res

    async def _assign(self, assign_handler: AssignHandler, args, kwargs):
        try:
            result = await self.loop.run_in_executor(self.threadpool, self._assign_threaded, args, kwargs, self.loop, assign_handler, self.provide_handler)
            shrinked_returns = await shrinkOutputs(self.template.node, result) if self.shrinkOutputs else result
            await assign_handler.pass_result(shrinked_returns)
        except Exception as e:
            await assign_handler.pass_exception(e)


from asyncio.futures import Future
from bergen.messages.postman.reserve.reserve_transition import ReserveState
from bergen import messages
from bergen.contracts.exceptions import AssignmentException
from bergen.registries.client import get_current_client
from bergen.schema import Node, NodeType
from bergen.monitor import Monitor, current_monitor
from bergen.messages.postman.reserve.params import ReserveParams
from bergen.messages import *
from bergen.utils import *
from rich.table import Table
from rich.panel import Panel
import asyncio
import logging


logger = logging.getLogger(__name__)


class Reservation:

    def __init__(self, node: Node, provision: str = None, loop=None, monitor: Monitor = None, ignore_node_exceptions=False, enter_on=[ReserveState.ACTIVE], exit_on=[ReserveState.ERROR, ReserveState.CANCELLED], bounced=None, **params) -> None:
        bergen = get_current_client()
        self._postman = bergen.getPostman()

        self.node = node
        self.params = ReserveParams(**params)
        self.monitor = monitor or current_monitor.get()
        self.ignore_node_exceptions = ignore_node_exceptions
        self.state_hook = None
        self.exit_states = exit_on
        self.enter_states = enter_on
        self.provision = provision

        self.bounced = bounced # with_bounced allows us forward bounced checks

        if self.bounced:
            assert "can_forward_bounce" in bergen.auth.scopes, "In order to use with_bounced forwarding you need to have the can_forward_bounced scope"

        if self.monitor:
            self.monitor.addRow(self.build_panel())
            self.log = lambda level, message: self.table.add_row(level, message)
            self.on_progress = lambda message, level: self.log(f"[magenta]{level}", message) if self.monitor.progress else None
        else:
            self.log = lambda level, message: logger.info(message)
            self.on_progress = None


        self.in_sync = False
        self.loop = loop or asyncio.get_event_loop()
        # Status
        self.running = False
        self.reservation = None
        self.critical_error = None
        self.recovering = False #TODO: Implement

        pass
    
    def build_panel(self):
        heading_information = Table.grid(expand=True)
        heading_information.add_column()
        heading_information.add_column(style="green")

        reserving_table = Table(title=f"[bold green]Reserving on ...", show_header=False)
        for key, value in self.params.dict().items():
            reserving_table.add_row(key, str(value))

        heading_information.add_row(self.node.__rich__(), reserving_table)

        self.table = Table()
        self.table.add_column("Level")
        self.table.add_column("Message")

        columns = Table.grid(expand=True)
        columns.add_column()

        columns.add_row(heading_information)
        columns.add_row(self.table)

        return Panel(columns, title="Reservation")


    async def assign_async(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, **kwargs):
        assert self.node.type == NodeType.FUNCTION, "You cannot assign to a Generator Node, use the stream Method!"
        self.log("[green]ASSIGN",f"Assigning!")
        if self.critical_error is not None:
            
            self.log("[red]ASSIGN",f"Contract is broken and we can not assign. Exiting!")
        try:
            shrinked_args, shrinked_kwargs = await shrinkInputs(self.node, args, kwargs) if not bypass_shrink else (args, kwargs)
            return_message = await self._postman.assign(self.reservation, shrinked_args, shrinked_kwargs=shrinked_kwargs, on_progress=self.on_progress, persist=persist, bounced=self.bounced)
            outs = await expandOutputs(self.node, return_message.data.returns) if not bypass_expand else return_message.data.returns
            return outs

        except AssignmentException as e:
            self.log("[red]ASSIGN", str(e))
            if not self.ignore_node_exceptions: raise e
        except Exception as e:
            raise e


    



    async def stream(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, **kwargs):
        assert self.node.type == NodeType.GENERATOR, "You cannot stream a Function Node, use the assign Method!"
        if self.critical_error is not None:
            self.log("[red]ASSIGN",f"Contract is broken and we can not assign. Exiting!")
        try:
            shrinked_args, shrinked_kwargs = await shrinkInputs(self.node, args, kwargs) if not bypass_shrink else (args, kwargs) 
            async for message in self._postman.assign_stream(self.reservation, shrinked_args, serialized_kwargs=shrinked_kwargs, with_progress=True, persist=persist, bounced=self.bounced):

                if isinstance(message, AssignYieldsMessage):
                    outs = await expandOutputs(self.node, message.data.returns) if not bypass_expand else message.data.returns
                    yield outs

                if isinstance(message, AssignLogMessage):
                    if self.on_progress: self.on_progress(message.data.message, message.data.level)
                
                if isinstance(message, AssignCriticalMessage):
                    raise AssignmentException(message.data.type + message.data.message)

                if isinstance(message, AssignDoneMessage):
                    self.log("ASSIGN", f'Done')
                    break

        except asyncio.CancelledError as e:
            self.log("[red] Cancelled")
            raise e

        except AssignmentException as e:
            self.log("[red]ASSIGN", str(e))
            if not self.ignore_node_exceptions: raise e

        
    

    async def contract_worker(self):
        self.running = True
        try:
            async for message in self._postman.reserve_stream(node_id=self.node.id, provision=self.provision, params_dict=self.params.dict(), with_progress=True, bounced=self.bounced):
                # Before here because Reserve Critical is actually an ExceptionMessage
                #TODO: Undo this

                if isinstance(message, ReserveLogMessage):
                    self.log(f'[green]{message.data.level.value}', message.data.message)

                elif isinstance(message, ReserveCriticalMessage):
                    # Reserve Errors are Errors that happen during the Reservation
                    self.log(f'[red]EXCEPTION', message.data.message)
                    self.critical_error = message

                elif isinstance(message, ExceptionMessage):
                    # Porotocol Exceptions are happening on the start
                    self.contract_started.set_exception(message.toException())
                    return

                elif isinstance(message, ReserveTransitionMessage):
                    # Once we acquire a reserved resource our contract (the inner part of the context can start)
                    state = message.data.state

                    if self.state_hook: await self.state_hook(self, message.data.state)
                    
                    if state in self.exit_states:
                        self.critical_error = f"Exited because Reservation Changed to State {state}: {message.data.message}"

                    if state in self.enter_states:
                        pass#self.contract_started.set_result(message.meta.reference)


                #elif isinstance(message, ReserveDoneMessage):
                    # Once we acquire a reserved resource our contract (the inner part of the context can start)
                #    self.contract_started.set_result(message.meta.reference)

                elif isinstance(message, ReserveActiveMessage):
                    # Once we acquire a reserved resource our contract (the inner part of the context can start)
                    self.contract_started.set_result(message.meta.reference)
        
        except asyncio.CancelledError as e:
            self.log("[green]DONE", "Unreserved Sucessfully")

 
    def cancel_reservation(self, future: Future):
        if future.exception():
            self.log("[red]Exception", str(future.exception()))
            raise future.exception()
        elif future.done():
            return


    async def start(self, state_hook = None):
        self.state_hook = state_hook
        return await self.__aenter__()

    async def end(self):
        await self.__aexit__()

    async def __aenter__(self):
        self.contract_started = self.loop.create_future()
        self.worker_task = self.loop.create_task(self.contract_worker())

        self.worker_task.add_done_callback(self.cancel_reservation)
        self.reservation = await self.contract_started
        self.log(f"[green]STARTED",f"Established Reservation {self.reservation}")
        return self

    async def __aexit__(self, *args, **kwargs):
        if not self.worker_task.done():
            #await self._postman.unreserve(reservation=self.reservation, on_progress=self.on_progress)
            self.worker_task.cancel()
            try:
                await self.worker_task
            except asyncio.CancelledError:
                self.log("[green]EXIT", "Gently Exiting Reservation")
            except Exception as e:
                self.log(f"[red]CRITICAL", f"Exitigin with {str(e)}")
    
    def assign(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, **kwargs):
        if self.in_sync:
            return self.loop.run_until_complete(self.assign_async(*args, bypass_shrink=False, bypass_expand=False, persist=persist, **kwargs))
        else:
            return self.assign_async(*args, bypass_shrink=False, bypass_expand=False, persist=persist, **kwargs)

    def __enter__(self):
        self.in_sync = True
        future = self.loop.run_until_complete(self.__aenter__())
        return future

    def __exit__(self, *args, **kwargs):
        return  self.loop.run_until_complete(self.__aexit__(*args, **kwargs))



    




       


import re
from typing import List, Tuple
from bergen.types.model import ArnheimModel
from bergen.registries.matcher import get_current_matcher
from bergen.enums import PortTypes
from bergen.schema import Node
import logging

logger = logging.getLogger(__name__)

class ExpansionError(Exception):
    pass



async def expandInputs(node: Node, args: list, kwargs: dict) -> dict:

    #assert node.inputs is not None, "Your Query for Nodes seems to not provide any field for inputs, please use that in your get statement"
    #assert len(node.inputs) > 0  is not None, "Your Node seems to not provide any inputs, calling is redundant"

    expanded_args = []
    for arg, port in zip(args, node.args):
        if port.TYPENAME == PortTypes.MODEL_ARG_PORT:
            modelClass = get_current_matcher().getModelForIdentifier(identifier=port.identifier)
            instance =  await modelClass.asyncs.get(id=arg)
            expanded_args.append(instance)
        else:
            expanded_args.append(arg)


    expanded_kwargs = {}
    for port in node.kwargs:
        if port.key not in kwargs:
            if port.required:
                raise ExpansionError(f"We couldn't expand {port.key} because it wasn't provided by our Kwargs, wrong assignation!!!")
            else:
                break

        if port.TYPENAME == PortTypes.MODEL_KWARG_PORT:
            modelClass = get_current_matcher().getModelForIdentifier(identifier=port.identifier)
            instance =  await modelClass.asyncs.get(id=kwargs[port.key])
            expanded_kwargs[port.key] = instance
        else:
            expanded_kwargs[port.key] = kwargs[port.key]


    return expanded_args, expanded_kwargs



async def shrinkOutputs(node: Node, returns: list) -> list:

    #assert node.inputs is not None, "Your Query for Nodes seems to not provide any field for inputs, please use that in your get statement"
    #assert len(node.inputs) > 0  is not None, "Your Node seems to not provide any inputs, calling is redundant"

    shrank_returns = []
    if not isinstance(returns, tuple) or isinstance(returns, list):
        returns = [returns]

    assert len(node.returns) == len(returns), "Returns do not conform to Node definition"
    for port, item in zip(node.returns, returns):
        if port.TYPENAME == PortTypes.MODEL_RETURN_PORT:
            if isinstance(item, ArnheimModel):
                shrank_returns.append(item.id)
            else:
                shrank_returns.append(item)
        else:
            shrank_returns.append(item)

    return shrank_returns



async def shrinkInputs(node: Node, args: list, kwargs: dict) -> Tuple[dict, dict]:

    #assert node.inputs is not None, "Your Query for Nodes seems to not provide any field for inputs, please use that in your get statement"
    #assert len(node.inputs) > 0  is not None, "Your Node seems to not provide any inputs, calling is redundant"

    shrinked_args = []
    for arg, port in zip(args, node.args):
        if port.TYPENAME == PortTypes.MODEL_ARG_PORT:
            instance = arg
            if isinstance(instance, ArnheimModel):
                shrinked_args.append(arg.id)
            else:
                raise Exception("You didnt provide a model")
        else:
            shrinked_args.append(arg)

    shrinked_kwargs = {}

    for port in node.kwargs:
        if port.key not in kwargs:
            break

        if port.TYPENAME == PortTypes.MODEL_KWARG_PORT:
            instance = kwargs[port.key]
            if isinstance(instance, ArnheimModel):
                shrinked_kwargs[port.key] = kwargs[port.key].id
            else:
                shrinked_kwargs[port.key] = kwargs[port.key]
        else:
            shrinked_kwargs[port.key] = kwargs[port.key]



    
    return shrinked_args, shrinked_kwargs


async def expandOutputs(node: Node, returns: list, strict=False) -> List:

    #assert node.inputs is not None, "Your Query for Nodes seems to not provide any field for inputs, please use that in your get statement"
    #assert len(node.inputs) > 0  is not None, "Your Node seems to not provide any inputs, calling is redundant"

    expanded_returns = []
    assert len(node.returns) == len(returns), "Returns do not conform to Node definition. Someone might have intercepted a Request"
    for port, item in zip(node.returns, returns):
        if port.TYPENAME == PortTypes.MODEL_RETURN_PORT:
            try:
                modelClass = get_current_matcher().getModelForIdentifier(identifier=port.identifier)
                instance =  await modelClass.asyncs.get(id=item)
                expanded_returns.append(instance)
            except AssertionError as e:
                if strict: raise e
                logger.error(f"Couldn't expand outputs make sure to import the schema for {port.identifier}")
                expanded_returns.append(item)
        else:
            expanded_returns.append(item)

    if len(expanded_returns) == 1:
        # Single outputs are returned like that
        return expanded_returns[0]
    
    else:
        return expanded_returns








import logging
import re
from typing import Generator, Generic, List, Type, TypeVar

logger = logging.getLogger(__name__)


gqlparsed_with_variables = re.compile(r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)\((?P<arguments>[^\)]*)\)[\s]*{[\s]*(?P<firstchild>[^\(:]*).*")
gqlparser_without_variables = re.compile(r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)[\s]*{[\s]*(?P<firstchild>[^\(\{\s:]*).*")

class GQLException(Exception):
    pass


MyType = TypeVar("MyType")


class GQL(object):

    def __init__(self, query: str) -> None:
        self.query = query
        self.variables = None
        self._type = None
        self.m = gqlparsed_with_variables.match(self.query)
        self.has_variables = True
        if not self.m:
            self.m = gqlparser_without_variables.match(self.query)
            self.has_variables = False
            if not self.m:
                raise GQLException("Illformed request")

    def combine(self, variables: dict):
        self.variables = variables
        return {"query": self.query, "variables": self.variables}
        
    def parsed(self):
        assert self.variables and self.query, "Please specify query and set variables before parsing"
        return {"query": self.query, "variables": self.variables}

    @property
    def firstchild(self):
        return self.m.group("firstchild")

    @property
    def operation_name(self):
        return self.m.group("operation")

    @property
    def type(self):
        return self.m.group("type")

    def extract(self, result: dict):
        return result[self.firstchild]



class ListQuery(list):


    def __init__(self, *args, **kwargs):
        self._df = None
        super().__init__(*args, **kwargs)


    def to_df(self, exclude_unset=True, **kwargs):
        import pandas as pd
        if self._df is None: self._df = pd.json_normalize([rec.dict(exclude_unset=exclude_unset, **kwargs) for rec in self])
        return self._df

    def _repr_html_(self) -> str:
        return self.to_df().to_html()


    
class TypedGQL(GQL, Generic[MyType]):

    def __init__(self, query: str, cls: Type[MyType], aslist=False):
        self.aslist = aslist
        self.query = query
        self._cls = cls
        super().__init__(query)

    @property
    def cls(self) -> MyType:
        return self._cls

    async def run(self, ward=None, variables=None, **kwargs) -> MyType:
        ward = ward or self._cls.get_ward()
        returnedobject = await ward.run(self, variables=variables, **kwargs)
        assert returnedobject is not None, "We received nothing back from the Server! Refine your Query!"
        if isinstance(returnedobject,list): return ListQuery([self.cls(**item) for item in returnedobject])
        return self.cls(**returnedobject)

    def subscribe(self, ward=None, **kwargs) -> Generator[MyType, None,None]:
        ward = ward or self._cls.get_ward()
        return ward.subscribe(self, **kwargs)

    def _repr_html_(self) -> str:
        string = "Query <br/><table>"
        string + "</table>"
        return string


class Query(GQL, Generic[MyType]):

    def __init__(self, query: str, cls: Type[MyType], aslist=False):
        self.aslist = aslist
        self.query = query
        self._cls = cls
        super().__init__(query)

    @property
    def cls(self) -> MyType:
        return self._cls

    def run(self, ward=None, variables=None, **kwargs) -> MyType:
        ward = ward or self._cls.get_ward()
        returnedobject = ward.run(self, variables=variables,**kwargs)
        if isinstance(returnedobject,list): raise Exception("Received a freaking list..., Please run QueryList")
        if not isinstance(returnedobject, dict):
            return self.cls(returnedobject) if returnedobject else None
        return self.cls(**returnedobject) if returnedobject else None

    def subscribe(self, ward=None, **kwargs) -> Generator[MyType, None,None]:
        ward = ward or self._cls.get_ward()
        return ward.subscribe(self, **kwargs)

    def _repr_html_(self) -> str:
        string = "Query <br/><table>"
        string + "</table>"
        return string


class QueryList(GQL, Generic[MyType]):

    def __init__(self, query: str, cls: Type[MyType], aslist=False):
        self.aslist = aslist
        self.query = query
        self._cls = cls
        super().__init__(query)

    @property
    def cls(self) -> MyType:
        return self._cls

    def run(self, ward=None, variables=None, **kwargs) -> List[MyType]:
        ward = ward or self._cls.get_ward()
        returnedobject = ward.run(self, variables=variables, **kwargs)
        return ListQuery([self.cls(**item) for item in returnedobject]) if returnedobject else ListQuery([])

    def subscribe(self, ward=None, **kwargs) -> Generator[List[MyType], None,None]:
        ward = ward or self._cls.get_ward()
        return ward.subscribe(self, **kwargs)

    def _repr_html_(self) -> str:
        string = "Query <br/><table>"
        string + "</table>"
        return string

class AsyncQuery(GQL, Generic[MyType]):

    def __init__(self, query: str, cls: Type[MyType], aslist=False):
        self.aslist = aslist
        self.query = query
        self._cls = cls
        super().__init__(query)

    @property
    def cls(self) -> MyType:
        return self._cls

    async def run(self, ward=None, variables=None, **kwargs) -> MyType:
        ward = ward or self._cls.get_ward()
        returnedobject = await ward.run_async(self, variables=variables,**kwargs)
        if isinstance(returnedobject,list): raise Exception("Received a freaking list..., Please run QueryList")
        if not isinstance(returnedobject, dict):
            return self.cls(returnedobject) if returnedobject else None
        return self.cls(**returnedobject) 

    def subscribe(self, ward=None, **kwargs) -> Generator[MyType, None,None]:
        ward = ward or self._cls.get_ward()
        return ward.subscribe(self, **kwargs)

    def _repr_html_(self) -> str:
        string = "Query <br/><table>"
        string + "</table>"
        return string

class AsyncQueryList(GQL, Generic[MyType]):

    def __init__(self, query: str, cls: Type[MyType], aslist=False):
        self.aslist = aslist
        self.query = query
        self._cls = cls
        super().__init__(query)

    @property
    def cls(self) -> MyType:
        return self._cls

    async def run(self, ward=None, variables=None, **kwargs) -> List[MyType]:
        ward = ward or self._cls.get_ward()
        returnedobject = await ward.run_async(self, variables=variables,**kwargs)
        return ListQuery([self.cls(**item) for item in returnedobject]) if returnedobject else ListQuery([])

    def subscribe(self, ward=None, **kwargs) -> Generator[MyType, None,None]:
        ward = ward or self._cls.get_ward()
        return ward.subscribe(self, **kwargs)

    def _repr_html_(self) -> str:
        string = "Query <br/><table>"
        string + "</table>"
        return string

def DelayedGQL(gqlstring):
    return lambda model : TypedGQL(gqlstring, model)







from .assign import AssignMessage
from .assign_cancelled import AssignCancelledMessage
from .assign_critical import AssignCriticalMessage
from .assign_log import AssignLogMessage
from .assign_return import AssignReturnMessage
from .bounced_assign import BouncedAssignMessage
from .assign_yield import AssignYieldsMessage
from .assign_done import AssignDoneMessage
from .bounced_forwarded_assign import BouncedForwardedAssignMessage


from abc import abstractmethod
from abc import ABC
import asyncio
from bergen.wards.base import GraphQLException, MainWard, TokenExpired
import aiohttp
import requests
from bergen.console import console
from bergen.wards.base import WardException
from bergen.query import  TypedGQL
from typing import TypeVar

T = TypeVar("T")


class BareMainWard(MainWard):

    def __init__(self, client, loop=None):
        super().__init__(client, loop=loop)
        self._graphql_endpoint = f"{self.protocol}://{self.host}:{self.port}/graphql"
        

    async def connect(self):
        self.async_session = aiohttp.ClientSession(headers=self._headers)

    async def pass_async(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = the_query.query
        try:
            async with self.async_session.post(self._graphql_endpoint, json={"query": query_node, "variables": variables}) as resp:

                if resp.status == 403:
                    raise TokenExpired("Token was Expired")
            
                result = await resp.json() 

                if "errors" in result:
                    raise  GraphQLException(f"Ward {self._graphql_endpoint}:" + str(result["errors"]))

                return the_query.extract(result["data"])

        except:
            console.print_exception(show_locals=True)
            raise 
            
        
    async def disconnect(self):
        await self.async_session.close()





from abc import abstractmethod
from bergen.messages.postman.log import LogLevel
from bergen.clients.base import BaseBergen
from bergen.messages.base import MessageModel
from bergen.hookable.base import Hookable, hookable
from bergen.actors.base import Actor
from bergen.actors.functional import *
from bergen.provider.utils import createNodeFromActor, createNodeFromFunction
from bergen.console import console
from pydantic.main import BaseModel
from bergen.constants import OFFER_GQL
import logging
import asyncio
from bergen.models import Node, Template
import inspect
from bergen.messages import *

logger = logging.getLogger()


class PodPolicy(BaseModel):
    type: str



class OneExlusivePodPolicy(PodPolicy):
    type: str = "one-exclusive"

class MultiplePodPolicy(PodPolicy):
    type: str = "multiple"



def isactor(type):
    try:
        if issubclass(type, Actor):
            return True
        else:
            return False
    except Exception as e:
        return False



class BaseProvider(Hookable):
    ''' Is a mixin for Our Bergen '''
    helperClass = None


    def __init__(self, client: BaseBergen, loop=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.client = client
        self.loop = loop or client.loop

        self.template_actorClass_map = {}

        self.message_queue = asyncio.Queue()
        self.provisions = {}
        
        
    def log(self, message):
        console.log(f"[blue] {message}")


    def template(self, node: Node, policy: PodPolicy = MultiplePodPolicy(), auto_provide=False, on_provide=None,  on_unprovide=None, bypass_shrink=False, bypass_expand=False, **implementation_details):
        console.print("[blue] Registered to Provide")
        console.print(node)

        def real_decorator(function_or_actor):
            assert callable(function_or_actor) or issubclass(function_or_actor, Actor), "Please only decorate functions or subclasses of Actor"
            # TODO: Check if function has same parameters as node

            template = Template.objects.update_or_create(**{
                "node": node.id,
                "params": implementation_details,
                "policy": policy.dict()
            })


            if isactor(function_or_actor):
                actorClass = function_or_actor
            else:
                is_coroutine = inspect.iscoroutinefunction(function_or_actor)
                is_asyncgen = inspect.isasyncgenfunction(function_or_actor)
                is_function = inspect.isfunction(function_or_actor)

                class_attributes = {"assign": staticmethod(function_or_actor), "expandInputs": not bypass_expand, "shrinkOutputs":  not bypass_shrink}


                if is_coroutine:
                    actorClass =  type(f"GeneratedActor{template.id}",(FunctionalFuncActor,), class_attributes)
                elif is_asyncgen:
                    actorClass =  type(f"GeneratedActor{template.id}",(FunctionalGenActor,), class_attributes)
                elif is_function:
                    actorClass = type(f"GeneratedActor{template.id}",(FunctionalThreadedFuncActor,), class_attributes)
                else:
                    raise Exception(f"Unknown type of function {function_or_actor}")
            

            self.register_actor(str(template.id), actorClass)

            return actorClass

        return real_decorator


    def enable(self, allow_empty_doc=False, widgets={}, **implementation_details):
        """Enables the decorating function as a node on the Arnheim, you will find it as
        @provider/

        Args:
            allow_empty_doc (bool, optional): Allow the enabled function to not have a documentation. Will automatically downlevel the Node Defaults to False.
            widgets (dict, optional): Enable special widgets for the parameters. Defaults to {}.
        """
        def real_decorator(function_or_actor):
            assert callable(function_or_actor) or issubclass(function_or_actor, Actor), "Please only decorate functions or subclasses of Actor"

            if isactor(function_or_actor):
                self.log("Already is Actor. Creating Node")
                node = createNodeFromActor(function_or_actor, allow_empty_doc=allow_empty_doc, widgets=widgets)
            else:
                self.log("Is Function. Creating Node and Wrapping")
                node = createNodeFromFunction(function_or_actor, allow_empty_doc=allow_empty_doc, widgets=widgets)

            # We pass this down to our truly template wrapper that takes the node and transforms it
            template_wrapper = self.template(node, **implementation_details)
            function = template_wrapper(function_or_actor)
            return function

        return real_decorator


    @abstractmethod
    async def connect(self) -> str:
        raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def disconnect(self) -> str:
        raise NotImplementedError("Please overwrite")


    @abstractmethod
    async def forward(self, message: MessageModel) -> None:
        raise NotImplementedError("Please overwrite")


    async def on_message(self, message: MessageModel):
        await self.message_queue.put(message)


    async def handle_bounced_provide(self, message: BouncedProvideMessage):
        try:
            await self.on_bounced_provide(message)

            progress = ProvideLogMessage(data={
            "level": LogLevel.INFO,
            "message": f"Pod Pending"
            }, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})

            await self.forward(progress)


        except Exception as e:
            logger.error(e)
            critical_error = ProvideCriticalMessage(data={
            "message": str(e),
            "type": e.__class__.__name__
            }, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})
            await self.forward(critical_error)
            raise e


    async def handle_bounced_unprovide(self, message: BouncedUnprovideMessage):
        try:
            await self.on_bounced_unprovide(message)

            progress = UnprovideLogMessage(data={
            "level": LogLevel.INFO,
            "message": f"Pod Unproviding",
            "provision": message.data.provision,
            }, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})

            await self.forward(progress)

        except Exception as e:
            logger.exception(e)
            critical_error = UnprovideCriticalMessage(data={
            "message": str(e),
            "type": str(e.__class__.__name__),
            "provision": message.data.provision,
            }, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})

            await self.forward(critical_error)


    @hookable("bounced_provide", overwritable=True)
    async def on_bounced_provide(self, message: BouncedProvideMessage):
        actorClass = await self.get_actorclass_for_template(message.data.template)
        self.log(f"Got provision request for {message.data.template} and will entertain {actorClass.__name__}")
        await self.client.entertainer.entertain(message, actorClass) # Run in parallel

    @hookable("bounced_unprovide", overwritable=True)
    async def on_bounced_unprovide(self, message: BouncedUnprovideMessage):
        self.log(f"Got unprovision. Sending to entertainer")
        await self.client.entertainer.unentertain(message)

    @hookable("get_actorclass_for_template", overwritable=True)
    async def get_actorclass_for_template(self, template_id):
        assert template_id in self.template_actorClass_map, f"We have no Actor stored in our list for Template with ID {template_id}"
        return self.template_actorClass_map[template_id]

    def register_actor(self, template_id: str, actorClass: Type[Actor]):
        assert template_id not in self.template_actorClass_map, f"We cannot register two Actors for the same template {template_id}"
        self.template_actorClass_map[template_id] = actorClass
        self.log(f"Registered Actor {actorClass.__name__} for Template {template_id}")

    async def provide_async(self):

        self.log(" ------------- AWAITING MESSAGES -----------")
        while True:
            message = await self.message_queue.get()
            
            if isinstance(message, BouncedProvideMessage):
                logger.info("Received Provide Request")
                assert message.data.template is not None, "Received Provision that had no Template???"
                await self.handle_bounced_provide(message)

            elif isinstance(message, BouncedUnprovideMessage):
                logger.info("Received Unprovide Request")
                assert message.data.provision is not None, "Received Unprovision that had no Provision???"
                await self.handle_bounced_unprovide(message)
            else: 
                self.log(f"Provider: [red] Received unknown message {message}")

            self.message_queue.task_done()

            
    def provide(self):
        if self.loop.is_running():
            raise Exception("Cannot do this in a running loop, please await or create task of provide_async()")
        else:
            self.loop.run_until_complete(self.provide_async())

        # we enter a never-ending loop that waits for data
        # and runs callbacks whenever necessary.
        





from ..log import LogDataModel
from enum import Enum
from pydantic.main import BaseModel
from ....messages.types import  RESERVE_TRANSITION
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = RESERVE_TRANSITION
    extensions: Optional[MetaExtensionsModel]


class ReserveState(str, Enum):
    # Start State
    #Start State
    ROUTING = "ROUTING"
    # Life States
    PROVIDING = "PROVIDING"
    WAITING = "WAITING"


    REROUTING = "REROUTING"
    CANCELING = "CANCELING"
    ACTIVE = "ACTIVE"

    # End States
    ERROR = "ERROR"
    ENDED = "ENDED"
    CANCELLED = "CANCELLED"
    CRITICAL = "CRITICAL"


class ReserveTransitionData(MessageDataModel):
    state: ReserveState
    message: Optional[str]


class ReserveTransitionMessage(MessageModel):
    data: ReserveTransitionData
    meta: MetaModel



from bergen.schemas.herre.types import Application, User
from bergen.config.types import HerreConfig
import json
import logging
import os
import shelve
from abc import ABC, abstractmethod
import requests
from oauthlib.oauth2 import RefreshTokenGrant

logger = logging.getLogger(__name__)

from bergen.console import console


class AuthError(Exception):
    pass



class BaseAuthBackend(ABC):


    def __init__(self, config: HerreConfig, token_url="o/token/", authorize_url="o/authorize/", check_endpoint="auth/", force_new_token=False) -> None:
        # Needs to be set for oauthlib
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0" if config.secure else "1"
        if not config.secure: console.log("Using Insecure Oauth2 Protocol.. Please only for local and debug deployments")
        self.config = config
        self.base_url = f'{"https" if config.secure else "http"}://{config.host}:{config.port}/'
        self.check_url = self.base_url + check_endpoint
        self.auth_url = self.base_url + authorize_url
        self.token_url = self.base_url + token_url
        self.scopes = config.scopes + ["introspection"]
        self.client_id = config.client_id
        self.client_secret = config.client_secret  
        self.force_new_token = force_new_token  

        self.scope = " ".join(self.scopes)

        # Lets check if we already have a local toke
        config_name = "token.db"
        run_path = os.path.abspath(os.getcwd())
        self.db_path = os.path.join(run_path, config_name)
        self.token = None

        super().__init__()


    @abstractmethod
    def fetchToken(self, loop=None) -> str:
        raise NotImplementedError("This is an abstract Class")

    @property
    def access_token(self) -> str:        
        return self.token["access_token"]


    def refetch(self):
        """Refetches the Tokens"""
        assert self.token is not None, "Cannot refetch Token if already fetched"

        if "refresh_token" in self.token:
            result = requests.post(self.token_url, {"grant_type": "refresh_token", "refresh_token": self.token["refresh_token"], "client_id": self.client_id, "client_secret": self.client_secret})
            self.token = result.json()
            return

        self.logout()
        self.login()


    def logout(self):
        self.token = None
        with shelve.open(self.db_path) as cfg:
            cfg['token'] = None


    def login(self, force_new=False):
        try:
            with shelve.open(self.db_path) as cfg:
                self.token = cfg['token']
        except KeyError:
            pass

        if not self.token or force_new or self.force_new_token:
            try:
                self.token = self.fetchToken()
                with shelve.open(self.db_path) as cfg:
                    cfg['token'] = self.token

            except ConnectionResetError:
                raise Exception(f"We couldn't connect to the Herre instance at {self.base_url}")     
            except:
                raise

        # Checking connection
        print("Reached here")
        response = requests.get(self.check_url, headers={"Authorization": f"Bearer {self.token['access_token']}"})

        if response.status_code != 200:
            # Our Initial Token was wrong, lets try to get a new one
            self.token = self.fetchToken()
            response = requests.get(self.check_url, headers={"Authorization": f"Bearer {self.token['access_token']}"})

        assert response.status_code == 200, "Was not able to get a valid token"


    def getToken(self, loop=None, force=False) -> str:
        if self._accesstoken is None:
            if force or self.token is None or self.force_new_token:

                try:
                    self.token = self.fetchToken()
                except ConnectionResetError:
                    raise Exception(f"We couldn't connect to the Herre instance at {self.base_url}")     
                except:
                    logger.error(f"Couldn't fetch Token with config {self.config}")
                    raise
                

                with shelve.open(self.db_path) as cfg:
                    cfg['token'] = self.token

            # Checking connection
            response = requests.get(self.check_url, headers={"Authorization": f"Bearer {self.token['access_token']}"})

            if response.status_code != 200:
                self.token = self.fetchToken()
                response = requests.get(self.check_url, headers={"Authorization": f"Bearer {self.token['access_token']}"})


            assert response.status_code == 200, "Was not able to get a valid token"

            self._application = Application(**json.loads(response.content))

            print(self.token)
            self._accesstoken = self.token["access_token"]
            self._refresh_token = self.token["refresh_token"] if "refresh_token" in self.token else None

        return self._accesstoken



from bergen.enums import ClientType
from bergen.clients.default import Bergen


class HostBergen(Bergen):
    pass
    
    def __init__(self, *args, scopes= [],**kwargs) -> None:
        super().__init__(*args, client_type = ClientType.HOST, scopes=scopes + ["host"], **kwargs)


from bergen.messages.postman.provide.provide_transition import ProvideState, ProvideTransitionMessage
from bergen.messages.postman.log import LogLevel
from bergen.messages.postman.provide.provide_log import ProvideLogMessage
from bergen.messages.postman.provide.provide_critical import ProvideCriticalMessage
from bergen.messages.postman.provide.provide_done import ProvideDoneMessage
from typing import Dict
from bergen.schema import Template
from bergen.messages.postman.assign.assign_done import AssignDoneMessage
from bergen.debugging import DebugLevel
from bergen.handlers.base import ContractHandler
from bergen.messages import BouncedProvideMessage
from bergen.console import console


class ProvideHandler(ContractHandler[BouncedProvideMessage]):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__( *args, **kwargs)
        self.current_state = None


    @property
    def reference(self) -> str:
        return self.message.meta.reference


    @property
    def template_id(self) -> str:
        return self.message.data.template

    @property
    def bounced(self):
        return self.message.meta.token

    @property
    def meta(self):
        return {"extensions": self.message.meta.extensions, "reference": self.message.meta.reference}

    @property
    def template_id(self) -> str:
        return self.message.data.template

    @property
    def context(self) -> Dict:
        if self.context_set is False:
            console.log("Context was never set, no sense in accessing it...")
            return {}
        return self.active_context

    async def log(self, message, level =LogLevel.INFO):
        console.log(message)
        await self.pass_log(str(message), level=level)   

    def set_context(self, context):
        self.context_set = True
        self.active_context = context

    async def set_state(self, state, message=None):
        if self.current_state != state:
            self.current_state = state
            await self.pass_transition(state, message)

    async def pass_transition(self, state: ProvideState, message=None):
        transition_message = ProvideTransitionMessage(data={"state": state, "message": message}, meta=self.meta)
        return await self.forward(transition_message)

    async def pass_done(self):
        provide_done = ProvideDoneMessage(data={"ok": True}, meta=self.meta)
        return await self.forward(provide_done)

    async def pass_log(self, message: str, level = LogLevel.INFO):
        provide_log = ProvideLogMessage(data={"message": message, "level": level.value}, meta=self.meta)
        return await self.forward(provide_log)

    async def get_template(self) -> Template:
        return await Template.asyncs.get(id=self.message.data.template)

    async def pass_exception(self, exception):
        critical_message = ProvideCriticalMessage(data={"message": str(exception), "type": exception.__class__.__name__}, meta=self.meta)
        return await self.forward(critical_message)


from bergen.wards.base import ServiceWard, TokenExpired, GraphQLException
from bergen.schema import WardSettings
import logging
from abc import ABC

from bergen.console import console
from bergen.query import GQL, TypedGQL
from gql.gql import gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.aiohttp import log as aiohttp_logger
from gql.transport.requests import RequestsHTTPTransport

aiohttp_logger.setLevel(logging.WARNING)


class AIOHttpGraphQLWard(ServiceWard):
    can_subscribe = False

    def __init__(self, client, settings: WardSettings, loop=None) -> None:
        super().__init__(client, settings, loop=loop)
        self._graphql_endpoint = f"{self.protocol}://{self.host}:{self.port}/graphql"

    async def connect(self):
        self.async_transport = AIOHTTPTransport(url=self._graphql_endpoint, headers=self._headers)
        await self.async_transport.connect()

    async def negotiate(self):
        query_node = gql("""
                mutation Negotiate {
                    negotiate
                }
            """)
        response = await self.async_transport.execute(query_node, variable_values={})
        return response.data["negotiate"]
    
    async def pass_async(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = gql(the_query.query)
        try:
            try:
                response = await self.async_transport.execute(query_node, variable_values=variables)
            except Exception as e:
                console.print_exception()
                raise TokenExpired(f"Token Expired {e}")
                
            if response.errors:
                raise GraphQLException(f"Ward {self._graphql_endpoint}:" + str(response.errors))
            
            return the_query.extract(response.data)

        except:
            console.print_exception(show_locals=True)
            raise 
            
        
    async def disconnect(self):
        await self.async_transport.close()
